import unittest
import numpy as np
import matplotlib.pyplot as plt
from dasy.synthesizers.gaussian import GaussianSynth
from dasy.synthesizers.uniform import UniformSynth
from dasy.labelers.classification.centroids import CentroidsLabeler
from dasy.labelers.regression.linear import LinearRegressionLabeler

class TestQuickstart(unittest.TestCase):
    def test_quickstart(self):
        plt.clf()
### dasy for classification
# 1. Define the problem
        dim = 2 # dimension of each input
        classes = 2 # number of classes
        n = 100 # number of data points
# 2. Create synthetic input data X
        synth = GaussianSynth(dim=dim) 
        X = synth.sample(n=n) # sample n dim-dimensional data points from Gaussian
# 3. Assign labels y
        labeler = CentroidsLabeler(classes=classes, dim=dim) # for 2 classes, this creates linearly separable labels
        y = labeler.assign(X)
# 4. Plot
        plt.scatter(X.T[0], X.T[1], c=y)
        plt.title('Synthetic Classification Problem')
        plt.tight_layout()
        plt.savefig('synthetic_classification')
        plt.clf()

### dasy for regression
# 1. Define the problem
        dim = 1 # dimension of each input
        n = 50 # number of data points
# 2. Create synthetic input data X
        synth = UniformSynth(dim=dim) 
        X = synth.sample(n=n) # sample n dim-dimensional data points from Uniform distribution
# 3. Assign continuous targets y
        labeler = LinearRegressionLabeler(dim=dim)
        y = labeler.assign(X)
# 4. Plot
        plt.scatter(X.T[0], y)
        plt.axline((0, labeler.b), slope=labeler.w, color='r') # plot the underlying line which generated the targets
        plt.title('Synthetic Regression Problem')
        plt.tight_layout()
        plt.savefig('synthetic_regression')



import unittest
import numpy as np
import matplotlib.pyplot as plt
from dasy.synthesizers.uniform import UniformSynth 
from dasy.synthesizers.gaussian import GaussianSynth 

class TestSynthesizers(unittest.TestCase):
    def test_uniform_synth(self):
        plt.clf()
        dim = 2
        n = 50

        low = -.5
        high = .5 
        synth = UniformSynth(dim=dim, low=low, high=high)
        data = synth.sample(n=n)
        # test boundaries
        for x in data:
            self.assertTrue(np.all(x < high))
            self.assertTrue(np.all(x > low))

        plt.scatter(data.T[0], data.T[1])
        plt.title('Uniform Data')
        plt.tight_layout()
        plt.savefig('uniform_data')

    def test_gaussian_synth(self):
        plt.clf()
        dim = 2
        n = 100

        synth = GaussianSynth(dim=dim)
        data = synth.sample(n=n)

        plt.scatter(data.T[0], data.T[1])
        plt.title('Gaussian Data')
        plt.tight_layout()
        plt.savefig('gaussian_data')




import numpy as np
from dasy.core import DataSynthesizer

#TODO: switch from numpy random to Generator

class GaussianSynth(DataSynthesizer):
    """
    Samples from a single multivariate Normal Distribution
    """
    def __init__(self, dim=None, mean=None, cov=None):
        """
        @param dim: input dimension 
        @param mean: dim-dimensional point
        @param cov: dim x dim covariance matrix
        """
        super().__init__(dim)
        self.mean = mean
        self.cov = cov
        # set defaults
        if mean is None:
            self.mean = np.zeros(dim)
        if cov is None:
            self.cov = np.eye(dim)
        if len(self.mean) != dim:
            raise ValueError(f'mean must have dimension = {dim} (is {len(self.mean)})')
        if len(self.cov) != dim:
            raise ValueError(f'cov must have dimension = {dim} x {dim} (is {self.cov.shape[0]} x {self.cov.shape[1]})')

    def sample(self, n=100):
        """
        See numpy.random.multivariate_normal for details on params
        We add simple defaults for mean and cov
        """
        return np.random.multivariate_normal(mean=self.mean, cov=self.cov, size=n)




import unittest
import numpy as np
import matplotlib.pyplot as plt
from dasy.synthesizers.gaussian import GaussianSynth
from dasy.labelers.classification.centroids import CentroidsLabeler
from dasy.plotting_utils import plot_bisectors

class TestClassificationLabelers(unittest.TestCase):
    def test_centroids_labeler(self):
        dim = 2
        plt.clf()
        #plt.gca().set_aspect('equal', adjustable='box')
        #fig, ax = plt.subplots()
        f, (ax1, ax2) = plt.subplots(1, 2, sharey=True)
        f.suptitle('Gaussian Data with Centroids Labeling', y=0.84)
        ax1.set_aspect('equal', adjustable='box')
        ax2.set_aspect('equal', adjustable='box')
        ax1.set_xlim(-2, 2)
        ax1.set_ylim(-2, 2)
        ax2.set_xlim(-2, 2)
        ax2.set_ylim(-2, 2)

        synth = GaussianSynth(dim=dim)
        data = synth.sample(n=100)
        #plt.scatter(data.T[0], data.T[1])
        #plt.show()

        labeler = CentroidsLabeler(classes=2, dim=dim)
        labels = labeler.assign(data)
        ax1.scatter(data.T[0], data.T[1], c=labels)
        ax1.scatter(labeler.centroids.T[0], labeler.centroids.T[1], c='r')
        ax1.plot(labeler.centroids.T[0], labeler.centroids.T[1], '--')
        plot_bisectors(labeler.centroids[0], labeler.centroids[1], ax=ax1)
        ax1.title.set_text('classes=2')
        #plt.show()

        labeler = CentroidsLabeler(classes=3, dim=dim)
        labels = labeler.assign(data)
        ax2.scatter(data.T[0], data.T[1], c=labels)
        ax2.scatter(labeler.centroids.T[0], labeler.centroids.T[1], c='r')
        #plt.plot(labeler.centroids.T[0], labeler.centroids.T[1], '--')
        plot_bisectors(labeler.centroids[0], labeler.centroids[1], ax=ax2)
        plot_bisectors(labeler.centroids[0], labeler.centroids[2], ax=ax2)
        plot_bisectors(labeler.centroids[1], labeler.centroids[2], ax=ax2)
        ax2.title.set_text('classes=3')
        plt.tight_layout()
        plt.savefig('centroids_labeler')




import matplotlib.pyplot as plt

def plot_bisectors(a, b, ax=plt):
    midpoint = (a + b) / 2
    slope = - (b[0] - a[0]) / (b[1] - a[1])
    #plt.plot([midpoint[0], midpoint[0] + 1], [midpoint[1], midpoint[1] + slope], '-')
    ax.axline((midpoint[0], midpoint[1]), slope=slope, linestyle='--')



import unittest
import matplotlib.pyplot as plt
from dasy.synthesizers.uniform import UniformSynth
from dasy.labelers.regression.linear import LinearRegressionLabeler

class TestRegressionLabelers(unittest.TestCase):
    def test_linear_labeler(self):
        dim = 1
        n = 50
        synth = UniformSynth(dim=dim)
        labeler = LinearRegressionLabeler(dim=dim)
        X = synth.sample(n=n)
        y = labeler.assign(X)
        #print(X.shape)
        #print(y.shape)

        plt.clf()
        plt.scatter(X.T[0], y)
        plt.axline((0, labeler.b), slope=labeler.w, color='r')
        plt.title('Uniform Data x with Linear y + noise')
        plt.tight_layout()
        plt.savefig('linear_labeler')




import numpy as np
from dasy.core import Labeler

class LinearRegressionLabeler(Labeler):
    """
    Label inputs x according to a linear function y=wx+b + e
    e is a noise term

    This is a probabilistic labeler, unless noise e=0. 
    """
    def __init__(self, dim=2, w=None, b=None, noise=np.random.normal):
        """
        @param dim
        @param w
        @param b
        @param noise: scalar function that takes a size argument, or None for no noise 
        """
        super().__init__()
        self.dim = dim
        if w is not None:
            self.w = w
        else:
            self.w = np.random.normal(size=dim)
        if b is not None:
            self.b = b
        else:
            self.b = 0
        if noise is not None:
            self.noise = noise
        else:
            self.noise = self.zero_function

    def assign(self, X):
        """
        @param X: matrix of inputs, where each row is an input vector x
        """
        # row-wise dot-product (w dot x for each x in X); add bias; add noise
        y = np.sum(self.w * X, axis=1) + self.b + self.noise(size=len(X)) #TODO: make it work for a single input?
        return y

    def zero_function():
        return 0
    



        



import numpy as np
from dasy.core import DataSynthesizer

#TODO: switch from numpy random to Generator

class UniformSynth(DataSynthesizer):
    """
    Samples from a single Uniform Distribution on an interval
    """
    def __init__(self, dim, low=-10., high=10.):
        super().__init__(dim)
        self.low = low
        self.high = high

    def sample(self, n=100):
        return np.random.uniform(low=self.low, high=self.high, size=(n, self.dim))

